/* wptClipImportDlg.cpp - WinPT key import dialog
 *	Copyright (C) 2001-2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"
#include "wptCrypto.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptImport.h"


/* Import the PGP key data from the clipboard.
   Return value: 0 on success. */
gpgme_error_t
gpgme_op_clip_import (gpgme_ctx_t ctx)
{
    gpgme_error_t err = 0;
    gpgme_data_t keydata = NULL;
    
    err = gpg_data_new_from_clipboard (&keydata, 0);
    if (!err)
	err = gpgme_op_import (ctx, keydata);

    gpgme_data_release (keydata);
    return err;
}


/* Load the import statistics dialog with the results from @res. */   
void
print_import_status (gpgme_import_result_t res)
{
    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_IMPORT_STAT, glob_hwnd,
		      import_status_dlg_proc, (LPARAM)res,
		      _("Key Import Statistics"), IDS_WINPT_IMPORT_STAT);
    if (res->imported || res->secret_imported)
        keycache_set_reload (1); /* XXX: update flag */
}


/* Dialog procedure for importing key clipboard data. */
BOOL CALLBACK
clip_import_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static int is_revcert;
    static int has_seckeys;
    int rc = 0;
    /*int id=0;*/
    gpgme_error_t err;
    gpgme_ctx_t ctx;
    gpgme_import_result_t res;
    /*gpgme_key_t *selkeys = NULL;*/
    
    switch( msg ) {
    case WM_INITDIALOG:
        SetWindowText (dlg, _("Key Import"));
        SetDlgItemText (dlg, IDC_IMPORT_DOIT, _("&Import"));

	rc = implist_build (&lv, GetDlgItem (dlg, IDC_IMPORT_KEYLIST));
	if (rc)
	    BUG (NULL);
	SetForegroundWindow (dlg);
	center_window (dlg, NULL);
	rc = implist_load (lv, NULL, &is_revcert, &has_seckeys);
        return TRUE;
        
    case WM_DESTROY:
        if( lv ) {
            implist_delete(lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_SYSCOMMAND:
        if( LOWORD (wparam) == SC_CLOSE )
            EndDialog( dlg, TRUE );
        return FALSE;
        
    case WM_COMMAND:
        switch( LOWORD( wparam ) ) {
        case IDCANCEL:
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDC_IMPORT_DOIT:
            if (has_seckeys > 0) {
		msg_box (dlg, _("Some of the imported keys are secret keys.\n\n"
				"The ownertrust values of these keys must be\n"
				"set manually via the Key Properties dialog."),
			 _("Import"), MB_INFO);
            }
	    SetForegroundWindow (dlg);
	    center_window (dlg, NULL);

	    /* XXX: support interactive and only import some keys. */
	    /*
	    int n;
	    char keyid[32];
	    n = listview_count_items (lv, 0);	    
	    if( n > 1 && listview_count_items( lv, 1 ) > 0 ) {
		selkeys = calloc (n+1, sizeof (gpgme_key_t));
		if (!selkeys)
		    BUG (NULL);
		for (id = 0; id < n; id++) {
		    if (listview_get_item_state (lv, id)) {
			listview_get_item_text (lv, id, 2, keyid, DIM (keyid)-1);
			// XXX: set the keyid??
			// ( selkeys, keyid+2 );
		    }
		}
	    }
	    */

	    err = gpgme_new (&ctx);
	    if (err)
		BUG (NULL);
            err = gpgme_op_clip_import (ctx);
            if (err) {
		msg_box (dlg, gpgme_strerror (err), _("Import"), MB_ERR);
		gpgme_release (ctx);
                return FALSE;
            }
	    res = gpgme_op_import_result (ctx);
	    print_import_status (res);
	    if (res->no_user_id > 0) {            
                msg_box (dlg, _("Key without a self signature was dectected!\n"
				"(This key is NOT usable for encryption, etc)\n"
				"\n"
				"Cannot import these key(s)."), _("Import"), MB_INFO);
            }
	    gpgme_release (ctx);
            EndDialog (dlg, TRUE);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}
